
/*
 * socketToolKit.c
 *
 * Initiate socket at GUI side. The client side is within the file
 * clientSide.c
#define GLIBC_21
 */

#define _GNU_SOURCE
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>

#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/un.h>
#include <strings.h>
#include <stdio.h>
#include <sys/time.h>

#define SOCKTYPE SOCK_DGRAM
#define	PROTOCOL			"udp"
/*
 * No longer going to use tcp sockets.
 *
#define SOCKTYPE SOCK_STREAM
#define	PROTOCOL			"tcp"
 */

#include "socketToolKit.h"
#include "slabmessages.h"

#ifndef SLAB_MAX_MTU
#define SLAB_MAX_MTU 1024
#endif

extern int log_entry();

char oscBuffer[SLAB_MAX_MTU];

void checkSLabSocket();
void checkOscilloscope();

int remote_socket_descriptor, osc_sock_desc, api_sock_desc;

struct  sockaddr_un local_socket_addr;

/*
 * Set up a socket for the client programmes to talk to mixslab GUI.
 * This is called from the serverSide.c routines.
 */
int
open_remote_socket(name, port, listens, reqsig)
char		*name; /* service name, must be known */
int 		listens; /* Parameter for the number of connections accepted */
{
	int namelength, result;
	struct	servent		*service, service_tmp;
	struct  sockaddr_in remote_socket_addr;

#ifdef NEW_DEBUG
	log_entry(GUI_DEBUG_EXT, "open_remote_socket(%s, %i, %i, %i)",
		name, port, listens, reqsig);
#endif
	/*
	 * Open a socket with minimal error handling.
	 */
	while((remote_socket_descriptor = socket(PF_INET, SOCKTYPE, IPPROTO_UDP))
		== -1)
	{
		switch (errno)
		{
			case EPROTONOSUPPORT :
					log_entry(GUI_DEBUG_EXT, "Unsupported protocol");
					break;
			case EMFILE :
					log_entry(GUI_DEBUG_EXT, "File table full");
					break;
			case ENFILE :
					log_entry(GUI_DEBUG_EXT, "System Filetable full!");
					break;
			case EACCES :
					log_entry(GUI_DEBUG_EXT, "Socket Create:perm denied");
					break;
			case ENOBUFS :
					log_entry(GUI_DEBUG_EXT, "Buffer space up");
					break;
			case EPROTOTYPE :
					log_entry(GUI_DEBUG_EXT, "Wrong protocol");
					break;
			default :
					log_entry(GUI_DEBUG_EXT, "Unknown error");
		}
	}
	/*
	 * Get info on the pwmgr port for binding socket.
	 */
	if (strcmp(name, "admin") == 0)
	{
		if ((service = getservbyname(name, PROTOCOL)) == NULL)
		{
			/*
			 * We want to open an administration socket, thus:
			 */
			log_entry(GUI_DEBUG_EXT, "Getservbyname failed for admin socket");
			log_entry(GUI_DEBUG_EXT, "Using default service , (5001)");
			service = &service_tmp;
			service->s_port = 5001;
			service->s_name = name;
		}
	} else if (strcmp(name, SERVICE) == 0) {
		if (port > 1024)
		{
			/*
			 * The requested service is not available, ie unknown. Get the 
			 * expected default for the slabControl
			 */
			log_entry(GUI_DEBUG_EXT, "Using configured: %s, (%i)", name, port);
			service = &service_tmp;
			service->s_port = port;
			service->s_name = name;
		} else if ((service = getservbyname(name, PROTOCOL)) == NULL) {
			/*
			 * The requested service is not available, ie unknown. Get the 
			 * expected default for the slabControl
			 */
			log_entry(GUI_DEBUG_EXT, "Getservbyname failed");
			log_entry(GUI_DEBUG_EXT, "Using default: slab, (%i)", 5000);
			service = &service_tmp;
			service->s_port = 5000;
			service->s_name = SERVICE;
		} else
			log_entry(GUI_DEBUG_EXT, "Getservbyname worked");
	} else {
		/*
		 * Assume we have another service (for example, the oscilloscope at the
		 * moment, but maybe more).
		 */
		log_entry(GUI_DEBUG_EXT, "Using provided: %s, (%i)", name, port);
		service = &service_tmp;
		service->s_port = port;
		service->s_name = name;
	}
	/*
	 * zero the struct (fixed size)
	 */
	bzero((char *) &remote_socket_addr, sizeof (remote_socket_addr));
	/*
	 * Setup the structure values for subroutines.
	 */
	remote_socket_addr.sin_family = AF_INET;
	remote_socket_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	remote_socket_addr.sin_port = htons(service->s_port);
	/*
	 * Bind the socket:
	 */
	log_entry(GUI_DEBUG_EXT, "Attempting bind");

	if ((bind(remote_socket_descriptor, (struct sockaddr *) &remote_socket_addr,
		sizeof(remote_socket_addr))) < 0) {
#ifdef NEW_DEBUG
		log_entry(GUI_DEBUG_EXT, "Can't bind socket to name");
#endif
		switch(errno) {
			case EBADF :
					log_entry(GUI_DEBUG_EXT, "Bad file descriptor");
					break;
			case EADDRINUSE :
					log_entry(GUI_DEBUG_EXT, "Addr in use");
					break;
			case ENAMETOOLONG :
					log_entry(GUI_DEBUG_EXT, "Name too long");
					break;
			case EISDIR :
					log_entry(GUI_DEBUG_EXT, "Null socket name");
					break;
			case EINVAL :
					log_entry(GUI_DEBUG_EXT, "socket already bound");
					break;
			case EACCES :
					log_entry(GUI_DEBUG_EXT, "socket protected");
					break;
			case EFAULT :
					log_entry(GUI_DEBUG_EXT, "Invalid name");
					break;
			default :
					log_entry(GUI_DEBUG_EXT, "Unknown error");
		}
		return(-1);
	}
	/*
	 * This is ugly, I want to use this routine to connect to a diversity of
	 * socket based routines, due to the introduction of the oscilloscope.
	 * It is basically a real kludge, I should redefine these operations.....
	 */
	if (port == 5001)
	{
		if (SOCKTYPE == SOCK_STREAM) {
			if ((result = listen(remote_socket_descriptor, listens)) < 0) {
				perror("Listen failed");
				log_entry(GUI_DEBUG_EXT, "Listen to socket failed: %i", result);
				return(-2);
			}
		}
		/*
		 * The only other operation I would put in here is to mark the socket as
		 * non-blocking, and to SIGUSR1 on receipt of data
		 */
		signal(SIGUSR1, checkOscilloscope);
		fcntl(remote_socket_descriptor, F_SETOWN, (long) getpid());
#ifdef GLIBC_21
		fcntl(remote_socket_descriptor, F_SETSIG, SIGUSR1);
		fcntl(remote_socket_descriptor, F_SETFL, O_NONBLOCK|O_ASYNC);
#else
		/*
			IF WE DO NOT HAVE LIBC 2.1 SUPPORT, CAN OSCILLOSCOPE?
		 * We need to organise some alternative polling schema is used. This
		 * should use SIGIO for all socket IO?
		 */
		fcntl(remote_socket_descriptor, F_SETFL, O_NONBLOCK);
#endif
		osc_sock_desc = remote_socket_descriptor;
	} else {
		/*
		 * This is going to be for the control API
		 */
		if (SOCKTYPE == SOCK_STREAM) {
			if ((result = listen(remote_socket_descriptor, listens)) < 0) {
				perror("Listen failed");
				log_entry(GUI_DEBUG_EXT, "Listen to socket failed: %i", result);
				return(-2);
			}
		}
		/*
		 * The only other operation I would put in here is to mark the socket as
		 * non-blocking, and to SIGUSR2 on receipt of data
		 */
		signal(SIGUSR2, checkSLabSocket);
		fcntl(remote_socket_descriptor, F_SETOWN, (long) getpid());
#ifdef GLIBC_21
		fcntl(remote_socket_descriptor, F_SETSIG, SIGUSR2);
		fcntl(remote_socket_descriptor, F_SETFL, O_NONBLOCK|O_ASYNC);
#else
		/*
		 * This will now use SIGIO, and we only support the control API, not
		 * the oscilloscope.
		 */
		fcntl(remote_socket_descriptor, F_SETFL, O_NONBLOCK);
#endif
		api_sock_desc = remote_socket_descriptor;
	}

	return(remote_socket_descriptor);
}

void checkOscilloscope()
{
	int rCnt;

#ifdef NEW_DEBUG
	log_entry(GUI_DEBUG_EXT, "checkOscilloscope");
#endif

#ifndef GLIBC_21
	signal(SIGUSR1, checkOscilloscope);
#endif
	/*
	 * Read data, put in buffer, call oscilloscope code.
	 */
	if ((rCnt = recv(osc_sock_desc, (char *) &oscBuffer, SLAB_MAX_MTU * 2, 0))
		> 0)
		newOscData(oscBuffer, rCnt);
	while ((rCnt = recv(osc_sock_desc, (char *) &oscBuffer,
		SLAB_MAX_MTU * 2, 0)) > 0)
	{
printf("recv %i\n", recv);
	}
}

void checkSLabSocket()
{
	slabmessage buffer;
	int recvCount;
	struct sockaddr from;
	int fromlen, doneSend = -1;

#ifndef GLIBC_21
	signal(SIGUSR2, SIG_IGN);
#endif
printf("checkSLabSocket\n");
	/*
	 * Since we are now going to use USR2 for multiple events, we need to 
	 * select on the available sockets.
	 */

	fromlen = sizeof(struct sockaddr);

	while ((recvCount =
		recvfrom(api_sock_desc, (char *) &buffer, 
		SLAB_MAX_MSG_SIZE, 0, &from, &fromlen)) > 0)
	{
		int res;

		printf("Received %i bytes\n", recvCount);

		if ((res = extractSLabMessage(&buffer, recvCount)) < 0)
		{
			printf("Packet Decode failed: %i\n", res);
			/*
			 * Return a frame anyway. We should do network byte conversions
			 * and also consider including some return status?
			 */
			sendto(api_sock_desc, (char *) &buffer, recvCount, 0,
				&from, fromlen);
			return;
		}
		/*
		 * We need to capture a few potention commands here. Diverse file
		 * operations need special handling, MIDI also. We may want to 
		 * introduce another entry into the interpreter array, where we
		 * place a name for a call the jumps into the mixslab binary.
		 */
		switch(buffer.command)
		{
			case SLAB_MIDI_OP:
				switch(buffer.message.midiMsg.operation)
				{
					case MIDI_CLOCK:
						/*
						 * We are going to have time sync put in here.
						 * This tacks into the GUI where the calculation
						 * of deltas are made for output sync.
printf("TIME SYNC: %i\n", buffer.message.midiMsg.time);
						 */
						doMidiSync(buffer.message.midiMsg.time);
						break;
				}
				break;
			default:
				doGUIoperation(
					buffer.command,
					buffer.message.trkMsg.track,
					buffer.message.trkMsg.controller,
					buffer.message.trkMsg.operator,
					buffer.message.trkMsg.value);
		}
		/*
		 * Return buffer as an ack. This should really be re-encoded to network
		 * byte ordering, etc.
		 */
		if (doneSend < 0)
		{
			sendto(api_sock_desc, (char *) &buffer, recvCount, 0,
				&from, fromlen);
			doneSend = 1;
		}
	}

#ifndef GLIBC_21
	signal(SIGUSR2, checkSLabSocket);
#endif
}

int
close_socket(sock_descriptor, message, length)
char *message;
{
#ifdef NEW_DEBUG
	log_entry(GUI_DEBUG_EXT, "Executing close_socket()");
#endif
	if (sock_descriptor >= 0)
	{
		if (length > 0)
			write(sock_descriptor, message, length);
		close(sock_descriptor);
	}
}

